#!python3
depends = [
    '受付期間',
    '申請期間',
    '申請受付',
    '申請受付期間'
    
    # get
    "申出期間",
    "申請の期間",
    "請求期間",
    "貸付期間",
    "応募受付期間",
    "申請時期",
    "貸付の期間",
    "貸与期間",
    "補助申請受付期間",
    "補助受付期間",
]



#!python3
from eyes_soatra import eyes
import requests
import json
from lxml.html.clean import Cleaner
from lxml import html
from lxml import etree
import re
from translate import Translator
import pandas

a = eyes.view_page(
    url='https://www.town.omachi.lg.jp'
)

print(a)


#!python3
from eyes_soatra.constant.depends.app_date.period import depends as __depends_period
from eyes_soatra.constant.depends.app_date.start import depends as __depends_start
from eyes_soatra.constant.depends.app_date.end import depends as __depends_end
from eyes_soatra.constant.depends.app_date.formats.period import formats as __formats_period
from eyes_soatra.constant.depends.app_date.formats.start import formats as __formats_start
from eyes_soatra.constant.depends.app_date.formats.end import formats as __formats_end
from eyes_soatra.constant.user.user_agents import User_Agents as __User_Agents
from eyes_soatra.funcs.utils.list import find as __find
from eyes_soatra.funcs.utils.list import map_list as __map_list
from eyes_soatra.funcs.utils.list import filter_list as __filter
from eyes_soatra.funcs.utils.string import strip_space as __strip_space
from eyes_soatra.funcs.utils.string import symbol as __symbol_all
from eyes_soatra.funcs.utils.string import xpath_tag as __xpath_tag
from eyes_soatra.constant.libs.requests import requests as __requests
from eyes_soatra.constant.vars import header_xpaths as __header_xpath
from eyes_soatra.constant.vars import description_xpath as __description_xpath
from eyes_soatra.constant.vars import full_stops as __full_stops
from eyes_soatra.constant.vars import remove_tags as __remove_tags
from eyes_soatra.constant.vars import priority_tag as __priority_tags
from eyes_soatra.constant.vars import priority_header_tag as __priority_header_tags
from eyes_soatra.constant import labels

from translate import Translator as __Translator
from lxml import html as __html
from lxml import etree as __etree

import jellyfish as __jellyfish
import random as __random
import time as __time
import re as __re
import json as __json

__separator = '\||-|:|\s+'
__header_min_length = 4
__date_max_length = 60

def __highlighter(
    html,
    xpath,
    xpath_desc,
    separator
):
    texts = []
    blogs = []
    separator = __separator + (separator if separator else '')
    
    xpaths = (__header_xpath + (xpath if type(xpath) == list else []))
    xpaths = __map_list(lambda each: f'({each})', xpaths)
    xpath = '|'.join(xpaths)
    
    xpaths_desc = (__description_xpath + (xpath_desc if type(xpath_desc) == list else []))
    tags_desc = __map_list(lambda xpath: __xpath_tag(xpath), xpaths_desc)
    
    elements = html.xpath(xpath)
    index = 0

    for element in elements:
        tag = element.tag
        temp_blogs = []
        text_list = element.xpath('.//text()')

        for text in text_list:
            for token in __re.split(separator, text):
                if token and not __symbol_all(token):
                    texts.append({
                        'tag': tag,
                        'token': token
                    })
                    temp_blogs.append({
                        'index': index,
                        'token': token
                    })
                    index += 1

        if temp_blogs and tag in tags_desc:
            blogs.append({
                'tag': tag,
                'blogs': temp_blogs
            })
    
    html_texts = []

    for text in html.xpath('//text()'):
        for token in __re.split(separator, text):
            if token and not __symbol_all(token):
                html_texts.append(token)
    
    return {
        'texts': texts,
        'blogs': blogs,
        'html-texts': html_texts
    }
    
def __type_formats(type):
    if type == labels.APP_START:
        return __formats_start
    
    if type == labels.APP_END:
        return __formats_end
    
    else:
        return __formats_period
    
def __best_point(word, depends):
    point = 0
    
    for depend in depends:
        temp = __jellyfish.jaro_similarity(word, depend)
        
        if point < temp:
            point  = temp

    return point
    
def __next_word(type_app, keyword, highlight):
    texts = highlight['html-texts']
    founds = []
    
    for i in range(0, len(texts)):
        if texts[i] == keyword and i < len(texts) - 1:
            next_word = texts[i + 1]
            
            if not next_word in founds:
                founds.append(next_word)

    if len(founds):
        formats = __type_formats(type_app)
        nexts = {
            'word': founds[0],
            'point': __best_point(founds[0], formats)
        }
        
        for i in range(1, len(founds)):
            point = __best_point(founds[i], formats)
            
            if point > nexts['point']:
                nexts = {
                    'word': founds[i],
                    'point': point
                }
        
        return nexts['word']

    return None

def __find_priority(founds):
    for tag in __priority_tags:
        found = __find(
            lambda each: each['tag'] == tag,
            founds
        )
        
        if found:
            return found
        
    priority = founds[0]
    
    for i in range(1, len(founds)):
        if len(priority['token']) > len(founds[i]['token']):
            priority = founds[i]
            
    return priority

def __specific_blog(
    next_word,
    keyword,
    highlight,
):
    texts = highlight['texts']
    texts_tokens = __map_list(lambda each: each['token'], texts)
    blogs = highlight['blogs']
    founds = []

    for each_blog in blogs:
        blogs_inner = each_blog['blogs']
        tokens = __map_list(lambda each: each['token'], blogs_inner)
        
        if next_word in tokens:
            for index in range(0, len(blogs_inner)):
                each = blogs_inner[index]
                
                if each['token'] == next_word:
                    preceding = each['index'] - 1
                    preceding = preceding if preceding >= 0 else 0
                    
                    if texts_tokens[preceding] == keyword:
                        str_token = ' '.join(tokens[index:])
                        found = __find(
                            lambda each: each['tag'] == each_blog['tag'],
                            founds
                        )
                        if found:
                            if len(found['token']) > len(str_token):
                                found['token'] = str_token

                        else:
                            founds.append({
                                'tag': each_blog['tag'],
                                'token': str_token
                            })
    
    priority = __find_priority(founds)
    if priority:
        return priority['token']

    return None

def __translate(lang, highlight):
    if not (lang == 'ja' or lang == 'en'):
        translate = __Translator(from_lang=lang, to_lang='en')
        
        highlight['html-texts'] = __map_list(
            lambda each: translate.translate(each),
            highlight['html-texts']
        )
        highlight['texts'] = __map_list(
            lambda each: {
                **each,
                'token': translate.translate(each['token'])
            },
            highlight['texts']
        )
        highlight['blogs'] = __map_list(
            lambda each: {
                **each,
                'blogs': __map_list(
                    lambda each: {
                        **each,
                        'token': translate.translate(each['token'])
                    },
                    each['blogs']
                )
            },
            highlight['blogs']
        )

    return highlight

def __header_best(founds):
    for found in founds:
        if found['tag'] in __priority_header_tags:
            return found
    
    header = founds[0]
    
    for i in range(1, len(founds)):
        if header['point'] < founds[i]['point']:
            header = founds[i]

    return header    

def __check_each(
    type_app,
    min_point,
    depends,
    highlight,
):
    texts = highlight['texts']
    result = {}
    point_temp = 0
    founds = []

    for text_obj in texts:
        token = text_obj['token']
        
        if len(token) >= __header_min_length:
            for depend in depends:
                point = __jellyfish.jaro_similarity(depend, token)
                
                if point > point_temp:
                    point_temp = point
                    result['keyword'] = token
                    result['similar-to'] = depend
                    result['point'] = round(point, 2)

                if point >= min_point:
                    founds.append({
                        'tag': text_obj['tag'],
                        'keyword': token,
                        'similar-to': depend,
                        'point': round(point, 2)
                    })

    if len(founds):
        best_header = __header_best(founds)
        result = {
            'ticked': True,
            **result,
            **best_header,
            'next-word': __next_word(
                type_app,
                best_header['keyword'],
                highlight
            )
        }
    
    return result

def __table_title(html, keyword, highlight):
    blogs = highlight['blogs']
    
    for blog in blogs:
        if (
            (blog['tag'] == 'th' or blog['tag'] == 'td') and
            keyword in __map_list(lambda each: each['token'], blog['blogs'])
        ):
            for table in html.xpath('//table'):
                tr1 = table.xpath('./thead/tr[1]')
                tr1 = tr1 if tr1 else table.xpath('./tbody/tr[1]')
                tr1 = tr1 if tr1 else table.xpath('./tr[1]')
                
                if len(tr1):
                    tr1 = tr1[0]
                    if tr1.xpath('./th') and tr1.xpath('./td'):
                        return None
                    
                    ts = tr1.xpath('./*[self::td or self::th]')
                    position = 0

                    for th in ts:
                        text = th.text_content()
                        position += 1
                        
                        if __re.sub(r'\s+', '', text) == keyword:
                            return {
                                'table': table,
                                'position': position
                            }

    return None

def __table_worker(html, keyword, highlight):
    table_obj = __table_title(html, keyword, highlight)
    
    if table_obj:
        table = table_obj['table']
        position = table_obj['position']
        tr_position = 1 if table.xpath(f'./thead') else 2
        tds = table.xpath(f'./tbody/tr[{tr_position}]/td[position()={position}]')
        tds = tds if tds else table.xpath(f'./tr[{tr_position}]/td[position()={position}]')

        if len(tds):
            return __strip_space(tds[0].text_content())
        
    else:
        for tr in html.xpath('//tr'):
            tds = tr.xpath('./td')
            
            for i in range(0, len(tds)):
                td = tds[i]
                
                if __re.sub(r'\s+', '', td.text_content()) == keyword:
                    texts = tr.xpath(f'./td[position()>{i+1}]//text()')
                    
                    return __strip_space(' '.join(texts))

    return None

def __detail_page(
    highlight,
    min_point,
    depends_end,
    depends_start,
    depends_period,
    show_all_detail,
    give_all
):
    result = {}
    depends = {
        labels.APP_START: (__depends_start + (depends_start if type(depends_start) == list else [])),
        labels.APP_END: (__depends_end + (depends_end if type(depends_end) == list else [])),
        labels.APP_PERIOD: (__depends_period + (depends_period if type(depends_period) == list else [])),
    }
    
    for key in depends:
        checked_result = __check_each(
            key,
            min_point,
            depends[key],
            highlight,
        )
        
        if show_all_detail:
            result[key] = checked_result

        elif 'ticked' in checked_result:
            result[key] = checked_result
            
    if not give_all:
        if labels.APP_PERIOD in result:
            for key in [labels.APP_START, labels.APP_END]:
                if key in result:
                    result.pop(key)

    return result

def __get_time(
    html,
    detail,
    highlight,
    give_all
):
    result = {}
    has_period = 'app-period' in detail and 'ticked' in detail['app-period']
    
    for key in detail:
        if not give_all and has_period:
            if key == 'app-start' or key == 'app-end':
                continue
            
        app = detail[key]
        
        if 'ticked' in app:
            keyword = app['keyword']
            text_td = __table_worker(
                html,
                keyword,
                highlight,
            )
            
            if text_td:
                result[key] = text_td
                
            else:
                next_word = app['next-word']

                if next_word:
                    specific = __specific_blog(
                        next_word,
                        keyword,
                        highlight
                    )

                    if specific:
                        result[key] = next_word
                        
                        for char in specific.removeprefix(next_word):
                            result[key] += char

                            if (
                                char in __full_stops or
                                (
                                    __re.search('\s', char) and
                                    len(result[key]) >= __date_max_length
                                )
                            ):
                                break

    return result

# ----------- public function
def time_app(
    url=None,
    content=None,
    lang='ja',
    xpath=None,
    xpath_desc=None,
    timeout=15,
    verify=False,
    headers=None,
    separator=None,
    sleep_time=2,
    tries_timeout=3,
    tries_reject=25,
    tries_forward=10,
    min_point=0.85,
    depends_end=None,
    depends_start=None,
    depends_period=None,
    allow_redirects=True,
    show_detail=True,
    show_all_detail=False,
    show_texts=False,
    show_blog=False,
    give_all=False,
    
    **requests_options
):
    if url == None and content == None:
        raise Exception('Please input one of those: url and content.')
    
    if url:
        url = __re.sub(r'\s', '', url)

    tried = 0
    agents = []
    redirected_forward = False

    while True:
        try:
            tried += 1
            
            if content:
                status_code = 200
                redirected = False
                current_url = url
                html = __html.fromstring(content)
                __etree.strip_elements(html, *__remove_tags)
                
            else:
                user_agent = __random.choice(__User_Agents)
                
                while user_agent in agents:
                    user_agent = __random.choice(__User_Agents)
                    
                agents.append(user_agent)
                    
                response = __requests.get(
                    **requests_options,
                    url=url,
                    timeout=timeout,
                    allow_redirects=allow_redirects,
                    verify=verify,
                    headers={
                        **(headers if headers else {}),
                        'USER-AGENT': user_agent,
                        'ACCEPT' : 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
                        'ACCEPT-ENCODING' : 'gzip, deflate, br',
                        'ACCEPT-LANGUAGE' : 'en-US,en;q=0.9,zh-CN;q=0.8,zh;q=0.7,km-KH;q=0.6,km;q=0.5,ja-JP;q=0.4,ja;q=0.3',
                        'REFERER' : 'https://www.google.com/'
                    },
                )
                status_code = response.status_code
                redirected = redirected_forward if redirected_forward else response.is_redirect
                current_url = response.url
                
                if status_code >= 400 and status_code <= 499:
                    return {
                        'error': f'Client error responses: {status_code}',
                        'status': status_code,
                        'redirected': redirected,
                        'url': current_url,
                        'tried': tried,
                    }
                    
                if status_code >= 500 and status_code <= 599:
                    return {
                        'error': f'Server error responses: {status_code}',
                        'status': status_code,
                        'redirected': redirected,
                        'url': current_url,
                        'tried': tried,
                    }
                
                html = __html.fromstring(response.content)
                __etree.strip_elements(html, *__remove_tags)
                
                if allow_redirects:
                    meta_refresh = html.xpath("//meta[translate(@http-equiv,'REFSH','refsh')='refresh']/@content")
                    
                    if len(meta_refresh):
                        if tried < tries_forward:
                            content_refresh = meta_refresh[0]
                            content_slices = content_refresh.split(';')
                            
                            if len(content_slices) > 1:
                                url_refresh = __strip_space(content_slices[1])
                                
                                if url_refresh.lower().startswith('url='):
                                    url_refresh = url_refresh[4:]
                                    
                                redirected_forward = True
                                url = url_refresh
                                continue

                        else:
                            return {
                                'error': f'Out of forwarding tries.',
                                'redirected': True,
                                'url': url,
                                'tried': tried
                            }

            highlight = __highlighter(
                html,
                xpath,
                xpath_desc,
                separator
            )
            highlight = __translate(
                lang,
                highlight
            )
            detail = __detail_page(
                highlight,
                min_point,
                depends_end,
                depends_start,
                depends_period,
                show_all_detail,
                give_all
            )
            time_result = __get_time(
                html,
                detail,
                highlight,
                give_all
            )
            
            return {
                **time_result,
                'url': current_url,
                'tried': tried,
                'status': status_code,
                'redirected': redirected,
                **({'detail': detail} if show_detail and detail else {}),
                **({'texts': highlight['texts']} if show_texts else {}),
                **({'blogs': highlight['blogs']} if show_blog else {}),
            }

        except Exception as error:                    
            if (
                type(error) == __requests.exceptions.ConnectionError or
                type(error) == __requests.exceptions.SSLError
            ):
                if tried >= tries_reject:
                    return {
                        'error': f'{error.__class__.__name__}: {error}',
                        'url': url,
                        'tried': tried
                    }
                    
                __time.sleep(sleep_time)
                
            else :
                if tried >= tries_timeout:
                    return {
                        'error': f'{error.__class__.__name__}: {error}',
                        'url': url,
                        'tried': tried
                    }



#!python3
from eyes_soatra import eyes
import json

res = eyes.time_app(
    url = 'https://www.city.tagajo.miyagi.jp:443/fukushi/kenko/fukushi/tyouikin.html'
)
js = json.dumps(
    res,
    ensure_ascii=False,
    indent=4
)

print(js)


formats = [
    "：令和5年2月9日（木） ～ 令和5年9月29日（金）※17時必着 （予算がなくなり次第、予告なく募集を終了します。） ・",
    "転入後3か月から1年以内（申請時には就業後3か月以上経過していること）",
    "工事請負契約又は売買契約締結の日から２ヶ月以内。",
    "大学の正規の最短修学年限",
    "県から交付された特定不妊治療費助成事業承認決定通知書の交付日から1年以内",
    "・施工業者との契約後速やかに（遅くても屋根工事完了前までに）申込書類を提出してください。 ・修繕、模様替えの場合は着工前一週間前までに申し込み書類を提出してください。",
    "令和2年4月1日から令和5年3月31日 （請求期間が過ぎると第十一回特別弔慰金を受けることができなくなりますので、ご注意ください）",
    "健康診査を受けた日から6か月以内",
    "令和5年4月3日（月）から令和6年3月29日（金）まで ※ 1頭の猫に対して申請できるのは1回までです。 ※ 補助金交付予定額が予算に達した時点で締め切ります。",
    "令和5年2月28日まで",
    "令和5 年4 月24日（月）～令和4年5月19日（金）",
    "令和５年４月１日から令和６年３月３１日まで ※ただし、予算に達した時点で申し込みを締め切りますので、ご注意ください ■問い合わせ先",
    "令和4年8月31日 17時15分まで",
    "令和4年6月22日 ～ 令和4年10月31日",
    "提出先",
    "受診をした月から、なるべく1年以内 ※ 申請できる権利は5年で消滅となります",
    "令和4年7月1日(金曜日)から令和5年2月28日(火曜日) (注意)郵送必着",
    "令和4年2月4日(金曜日)から令和4年9月30日(金曜日)まで",
    "令和2(2020)年4月1日～令和5(2023)年3月31日",
    "令和９年３月３１日まで",
    "検査日が令和4年4月1日～ 令和5年5月7日検査分 まで",
    "年度をまたいで治療継続中の方は、治療実施翌年度の6月から3月末の期間",
    "第 1 回申請受付期間：平成 23 年12 月 1 日から平成 24 年3 月21 日",
    "・・・ 金融機関との協議による",
    "物件登録者 初めて登録された日から2年以内",
    "申請期間：令和4年7月4日（月曜日）～令和4年9月30日（金曜日） 申請は、先着順に受付いたします。 ただし、申請額が予算額を超えた日に複数の申請があった場合は、その日の申請人全員を対象とした抽選を別途行います。",
    "清掃日又は11条検査受検日のいずれか遅い日の3か月後まで （遅れると申請ができません。）",
    "：5月8日（金）～8月31日(月） ※オンライン申請は5月11日（月）から受付",
    "平成25年2月8日（金）から平成25年2月21日（木）まで 意見受付期間 平成25年2月8日（金）から平成25年2月21日（木）まで",
    "災害等による損害を受けた日から1年間",
    "在学している学校の修業年限です。",
    "令和5年5月8日（月曜日）～令和5年12月15日（金曜日）",
    "災害を受けた日から30日以内",
    "新築住宅=最長3年間（最大30万円）",
    "受付期間：令和5年10月31日（火曜日）まで",
    "令和5年4月1日から令和6年3月11日まで",
    "令和４年５月１６日（月）から ※ただし、受け付けできる件数に限りがございますので、ご了承ください。",
    "令和5年4月1日から令和6年3月31日",
    "令和４年６月１日から令和４年１０月３１日 午前９時から午後５時まで",
    "随時",
    "1または2のいずれか早い日までです。",
    "令和2年4月1日～令和5年3月31日 (請求期間を過ぎると第十一回特別弔慰金を受けることができなくなりますので、ご注意ください。)",
    "6月1日から11月30日まで ※1組織申請受付期間内1回の申請になります。 ※申請書一式は毎年5月下旬に横須賀市に結成を届出た「自主防災組織」の町内会・自治会長に配布します。",
    "毎年4月1日～5月末（途中申請の場合は、翌年1月末まで随時受け付けますが、受付翌月からの援助となります。）",
    "治療を受けた月の属する年度内（4月1日から翌年3月31日まで） とします。",
    "平成30年４月から 申請は各年度の初回のみ必要で、以後は支給額がある場合に、指定振込先へ振り込みます。 初回申請以降は毎年７月に申請が必要です。",
    "随時",
    "随時受け付けています。 予算に達し次第終了となります。",
    "禁煙外来治療終了日から1年以内の申請に限ります",
    "申請時期：本市に転入し上記に掲げる、「移住元要件」「移住先要件」及び「地域の担い手としての役割に関する要件」を満たした後。",
    "○4月1日～ 先着順 予算枠に達ししだい終了となります。 ○補助金交付決定は、4月中旬以降の予定です。 （申請様式等は下方にあります。↓）",
    "2022年4月15日(金曜日)～2023年3月31日(金曜日)まで。",
    "奨学金の貸与が決まった年度の４月から在学する学校の正規の修学期間",
    "毎年 年度末 3月20日まで（休日の場合は前日）",
    "医学生 6年以内",
    "令和2年4月1日から令和5年3月31日",
    "令和6年3月29日（金曜日）まで（書類必着）",
    "高校入学前の中学3年時の3月 （※詳しい日程については、毎年、中学3年生を対象に配布する募集案内に記載します。）",
    "令和５(2023)年４月3日（月）から令和６(2024)年３月29日（金）まで",
    "令和5年5月1日（月）から令和6年2月23日（金）まで",
    "令和4年9月30日（金曜日） ※当日消印有効",
    "貸与を受けた年度の3月末日まで",
    "令和５年４月１日より補助金額が変更となります。",
    "4月1日から受け付けます。その年度の3月31日までに完了する事業が対象です。",
    "年度当初の認定の場合、在校生は1月～2月、新1年生は4月上旬までです。 年度途中に生活状態が悪化した場合は、随時受け付けます。",
    "令和4年12月1日から令和5年1月31日まで（消印有効）",
    "令和4年11月30日まで （受付は終了しました）",
    "令和２年４月１日から令和５年３月３１日まで （請求期間を過ぎると第十一回特別弔慰金を受けることができなくなりますので、ご注意ください。）",
    "開庁時間内 随時",
    "毎年3月～ 4月(詳しくは、市のHPでお知らせします。)",
    "令和2年4月1日から令和5年3月31日 （請求期間を過ぎると第十一回特別弔慰金を受けることができなくなりますので、ご注意ください）",
    "就職活動を行った日から6か月以内",
    "学校における正規の修学期間",
    "令和4年度分 は、 令和4年度末 に資格を取得したことになどより、 令和5年4月以後に納期限 が設定されているもの。",
    "申込書：申込み期限 住宅の引き渡しの概ね１０日前まで（住宅の引渡し時点で申込書の受理通知が発行されていること） 申請書：申請期限",
    "毎年7月分～次年6月分 毎年7月1日から、随時申請できます。",
    "奨励措置期間は課税年度から10年以内とする。",
    "令和5年5月22日(月曜日)から令和5年9月29日(金曜日)",
    "令和５年４月１７日から 令和５年１１月３０日まで",
    "治療の種類によって申請期間が異なります（年度内申請ではなくなりました）。 ＊やむを得ない理由で申請期限を過ぎてしまう場合は、あらかじめ保健相談センターへご連絡ください。",
    "原則3か月以内（状況に応じて3か月以内での延長貸付や再貸付が受けられる場合あり） 注 要相談",
    "令和4年3月31日まで",
    "令和2年4月1日から 令和5年3月31日 まで（期間を過ぎると請求ができませんのでご注意ください）",
    "令和７年３月31日まで(必着)",
    "令和2年5月25日(月)～令和2年9月25日(金) ※土日祝日を除く8時30分～17時15分 申請先 上士幌町役場保健福祉課福祉担当窓口(郵送申請も可)",
    "令和3年(2021年)4月1日（木曜日)〜令和4年(2022年)3月31日(木曜日)",
    "令和3年5月19日（水曜日）から令和4年2月28日（月曜日）まで",
    "最後の妊婦健康診査受診日から6カ月以内",
    "随時",
    "令和4年8月1日月曜日から",
    "・1年間の診療分について、診療期間終了後の 3か月以内 までに申請を行う。",
    "令和元年 ８月１日 （木）～ 令和２ 年２月３日（月）",
    "助成対象講座を修了した日から3か月以内",
    "毎年4月1日から3月31日まで",
    "令和4年7月中旬の賦課決定通知書到着後から令和5年3月31日（当日消印有効）"
]